# -*- coding: utf8 -*-
'''
*******************************
***  DB BASE MODULE
*******************************
$RCSfile: base_dbmod.py,v $
$Revision: 1361 $
$Author: j.holetschek $
$Date: 2014-04-10 17:08:23 +0200 (Do, 10. Apr 2014) $

The biocase.wrapper database base module to access datasources.
This is a base class for other dbmods that at least need to implement the connection method
if they are compliant to the Python DB API 2.0.'''

from string import join
from biocase.wrapper.errorclasses import DbModuleNotSpecifiedError, SQLunicodeGenerationError, SQLError
from biocase.wrapper.sql.select import sqlSelect
from biocase.wrapper.sql.operators import getWhereObjTables

import logging
log = logging.getLogger("pywrapper.dbmod")

# uses inherited ANSI SQL

# ---------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------
class DBBasemodule(sqlSelect):
    __version__ = '0.1'
    __dbmod__ = u'DB base module'

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def __repr__(self):
        return ("%s v%s" % (self.__class__.__dbmod__, self.__class__.__version__))

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def __init__(self, psfObj):
        # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _

        sqlSelect.__init__(self, psfObj)
        #
        # connect to DB

        self.conn = self.createDBconnection()


    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def __del__(self):
        # release db connection
        try:
            self.conn.close()
            log.info("successfully closed %s db connection." % self.__dbmod__)
        except:
            log.error("error when trying to close the %s db connection." % self.__dbmod__)

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def createDBconnection(self):
        raise DbModuleNotSpecifiedError

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def execSelect(self, selectList, whereObj=None, orderByList=[], limit=None, offset=None):
        '''execute a select statement.
        the selectList gives the wanted attribute/alias tuples,
        the optional whereObj the WHERE filter,
        the optional orderbyList the ORDER BY clause as attribute/alias tuples,
        the optional limit the LIMIT/TOP limitation as an integer.'''
        # get SQL string
        (sql, paras) = self.select(selectList, whereObj, limit=limit, orderByList=orderByList, offset=offset)
        # execute statement
        return self.execSQL(sql, paras)

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def fixAllIds(self, AllIds):
        return AllIds

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def execDistinct(self, selectList, whereObj=None):
        '''Select distinct statement. The default routine does a regular select and filters only unique records via Python.'''
        log.debug("Select Distinct List: " + str(selectList))
        # build SELECT clause
        select = "DISTINCT %s" % (join([self.joinDBAttrObj(dbt) for dbt in selectList], ', '))
        # get SQL string
        (sql, paras) = self.select(selectList, whereObj, selectReplacementString=select, orderByList=selectList)
        # execute statement
        return [r[0] for r in self.execSQL(sql, paras)]

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def execCount(self, DBAttrObj, whereObj=None):
        '''SELECT COUNT(x) statement. The default routine does a regular select and counts the records via Python.
        If the filter refers to the root table or is none, we can use a simple select count (because there will be only distinct primary keys);
        otherwise, we have to manually count distinct PKs (because count distinct is not supported by all DBMS)'''
        recCount = -1
        whereTables = getWhereObjTables(whereObj)
        if len(whereTables) == 0 or (len(whereTables) == 1 and whereTables[0] == DBAttrObj.table):
            log.info('Only root table queried; using SELECT COUNT to get number of records.')
            select = self.apply_template('count', [self.joinDBAttrObj(DBAttrObj)])
            (sql, paras) = self.select([DBAttrObj], whereObj, selectReplacementString=select)
            res = self.execSQL(sql, paras)
            if len(res) == 0:
                recCount = 0
            else:
                recCount = res[0][0]
        else:
            if self.templates.has_key('count-distinct'):
                log.info('Other tables that root table used in query; using count distinct to get number of records.')
                select = self.apply_template('count-distinct', [self.joinDBAttrObj(DBAttrObj)])
                (sql, paras) = self.select([DBAttrObj], whereObj, selectReplacementString=select)
                res = self.execSQL(sql, paras)
                if len(res) == 0:
                    recCount = 0
                else:
                    recCount = res[0][0]
            else:
                log.info('Other tables that root table used in query, count distinct not supported by DBMS; getting list of root table PKs and counting distinct values manually.')
                (sql, paras) = self.select([DBAttrObj], whereObj)
                PKs = set([rec[0] for rec in self.execSQL(sql, paras)])
                recCount = len(PKs)

        # return result
        return recCount

    # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def execSQL(self, sql, paras=None):
        # create cursor:
        self.c = self.conn.cursor()
        # execute SQL statement and return result
        log.debug("start DB query")
        try:
            log_sql = sql
            sql = sql.encode(self.psfObj.encoding)
            if paras is not None:
                try:
                    log.info("Executing SQL: '%s' with parameters: %s" % (sql, "'" + join([unicode(p) for p in paras], "', '") + "'"))
                except:
                    log.info("Executing SQL: '%s'" % (log_sql))
                    log.debug("Could not log used SQL parameters")
            else:
                log.info("Executing SQL: '%s'" % (log_sql))
        except:
            self.c.close()
            raise SQLunicodeGenerationError()
        # EXECUTE SQL
        try:
            if paras is None:
                self.c.execute(sql)
            else:
                self.c.execute(sql, paras)
                #self.c.execute(sql, ('%', 'Acalyptophis peronii'))
        except:
            raise SQLError(sql, paras)

        tmp = self.c.fetchall()
        log.debug("end DB query")
        log.info("Hits: " + str(len(tmp)))
        if log.isEnabledFor(logging.DEBUG):
            try:
                if type(tmp) == type('a'):
                    rsString = unicode(tmp, self.psfObj.encoding)
                else:
                    rsString = unicode(tmp)
                log.debug("Resulting Recordset:  " + rsString)
            except:
                log.debug("The Recordset Result could not be converted to a string")
        self.c.close()
        return tmp

# _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
    def execSQLnonUnicode(self, sql, paras=None):

        # create cursor:
        self.c = self.conn.cursor()
        # execute SQL statement and return result
        log.debug("start DB query")

        # EXECUTE SQL
        try:
            if paras is None:
                self.c.execute(sql)
            else:
                self.c.execute(sql, paras)
        except:
            raise SQLError(sql, paras)
        tmp = self.c.fetchall()
        log.debug("end DB query")
        log.debug("Hits: " + str(len(tmp)))
        if log.isEnabledFor(logging.DEBUG):
            try:
                if type(tmp) == type('a'):
                    rsString = tmp, self.psfObj.encoding
                else:
                    rsString = tmp
                log.debug("Resulting Recordset:  " + rsString)
            except:
                log.debug("The Recordset Result could not be converted to a string")
        self.c.close()

        return tmp

    def getTables(self):
        log.error("in base")
        return None

    def getColumns(self, table):
        log.error("in base")
        return None

    def verifyTables(self, tableList):
        if self.getTables() == None:
            return []
        errTables = []
        dbTables = [t.lower() for t in self.getTables()]
        for confTable in [ct.lower() for ct in tableList]:
            if not confTable in dbTables:
                errTables.append(confTable)
        return errTables

    def hasMetaInfo(self):
        return False
